//*******************************************************
//***
//***       Arithmetic Interpreter        Version 2.2
//***
//***                 C++ Version
//***
//*******************************************************

/*

Cette portion de code permet de reconnaitre des expressions mathematiques
rentrees sous forme de chaines de caracteres et ensuite d'evaluer ces ex-
pressions.

Les fonctions reconnues sont:
+ - * / ^
exp ln log abs sin cos tan asin acos atan sinh (ou sh) cosh (ou ch)
tanh (ou th)
L'interpreteur reconnait les nombres reels,entiers positifs ou negatifs.
L'interpreteur reconnait les variables x et y.

L'interpreteur sait evaluer des expressions avec redondance de parentheses.
Exemple: (1+(2)) donnera 3.

Les fonctions unaires (sin,cos,abs,....), si elles n'admettent qu'un seul 
argument (Exemple sin(x)) peuvent etre ecrite aussi sans parentheses (Exemple:
sinx).   ATTENTION: ecrire sinx+3 est equivalent a sin(x)+3.

Pour beneficier de cet interpreteur, il suffit d'inclure le fichier CInterpreter.h
dans le projet et de faire appel aux fonctions:

friend CPPNode * ConstructTree(char *,short,short &);
float ComputeTree(float,float,short &);

La fonction ConstructTree sert a construire l'arbre binaire representant l'expres-
sion saisie au clavier.Elle prend en arguments une chaine de caractere a interpreter
, le nombre de caracteres de cette chaine (signe '\0' non inclus) et un short qui 
representera le type d'erreur rencontree.

IMPORTANT: les lettres doivent etre en minuscule et il ne doit pas y'avoir d'espace
sous peine de renvoi d'un code d'erreur 1,2 ou 3.

exemple: 3*x+1 est correct. 3*X+1  non      -> erreur de type 3. 
                            3* x+1 non plus -> erreur de type 3.

La fonction ComputeTree prend en arguments les valeurs respectives des variables x et y
et un short qui representera le type d'erreur rencontree.

Statistiques:
	
Sur la fonction: 1/(x+0.2)*sin(x^2)*cos(x), on obtient ce resultat somme toute correct:
				
  86500 evaluations/seconde sur un PowerMac 7200/90
 660000 evaluations/seconde sur une SUN Ultra 5 Solaris			
Teste sur:

SUN Ultra 5:
SOLARIS                 g++ CInterpreter.cpp -lm

PowerMac G3:
MacOS X Server (RDR2)
MacOS 8.x               CodeWarrior IDE

Historique:

2.2:
  * Version Objective-C disponible.
  * Amelioration du code de la version C++
    - Correction de bugs sur l'operator =
    - Optimisation du code
    - Reformatage pour plus de clarete.
  * Mise a jour du code de la version C.

2.1:
  * Version Java disponible:
    # les operateurs mathematiques disponibles sont ceux de Java donc certains manquent.

  * Correction d'un bug sur le calcul de puissance (Version C++).

2.0: 
  * Portage en C++ et debut en Java

Version C++:
	 
  * Meilleure gestion memoire lors de la construction de l'expression.
  * Acceleration de certaines operations.

Version Java:
	 
  * Premiere version. Alors si il y'a un bug, c'est fort possible.

1.3b: ajoute les fonctions suivantes: (NON DISTRIBUEE)
  * reconnaissance du symbole 
  * ecriture formatee d'expression (a ameliorer)

1.2: corrige les bugs suivants:
  * erreur sur l'interpretation de fonctions unaires imbriquees telles que
    ln(exp(x)).
  * la fonction puissance (pour les puissances entieres).
    ajoute:
  * la verification de la chaine entree (voir code d'erreur)
  * la verification de l'existence de l'evaluation (voir code d'erreur)
     

1.1: corrige un bug au niveau de l'interpretation des fonctions du type:

      exp(-(x+y))

1.0: premiere version

Ce code peut etre distribue et modifie librement.

*/

#include "CInterpreter.h"


//...........................................................CPPNode::CPPNode

CPPNode::CPPNode(void) {
fg=fd=NULL;
}

//...........................................................CPPNode::CPPNode

CPPNode::CPPNode(short inOperator,float inValue,CPPNode * Fg,CPPNode * Fd) {
mOperator=inOperator;
mValue=inValue;
fg=Fg;
fd=Fd;
}

//...........................................................CPPNode::CPPNode

CPPNode::CPPNode(const CPPNode & inCPPNode) {
mOperator=inCPPNode.mOperator;
mValue=inCPPNode.mValue;
fg=inCPPNode.fg;
fd=inCPPNode.fd;
}

#pragma mark -

//........................................................ CopyPartialString

char * CopyPartialString(char * inString,short inBegin,short inEnd) {
char *  tString;
short tLength=inEnd-inBegin+1;

tString=new char[tLength+1]; // NewPtr ou calloc

if (tString!=NULL) {
  memmove(tString,inString+inBegin,tLength);
  tString[tLength]='\0';
}

return tString;
}

//............................................................... CalcValue

float CalcValue(char * inString,short inIndex) {
short i=inIndex,j;
float tValue=0;
short sign=1;

if (inString[inIndex]=='') return PI;
if (inString[inIndex]=='-') {
	sign=-1;
	i++;
}

while (inString[i]>47 && inString[i]<58) {
	tValue=tValue*10+(inString[i]-'0');
	i++;
}

if (inString[i]=='.') {
	i++;
	j=1;
	
  while (inString[i]>47 && inString[i]<58) {
		tValue=tValue+(inString[i]-'0')*exp(-j*2.30258509);
		i++;
		j++;
	}
}

return (sign*tValue);
}

//............................................................. FindOperator

void FindOperator(short * oper,short * esp,char * inString,short pos) {

switch(inString[pos]) {
	case 'a':
		switch(inString[pos+1]) {
			case 'b':
				*esp=3;
				*oper=10;
				break;
			case 'c':
				*esp=4;
				*oper=15;
				break;
			case 's':
				*esp=4;
				*oper=14;
				break;
			case 't':
				*esp=4;
				*oper=16;
				break;
		}
		break;
	case 'c':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=18;
		}
		else if (inString[pos+1]=='o' && inString[pos+2]=='s') {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=18;
			}
			else {
        *esp=3;
				*oper=12;
			}
		}
		break;
	case 'e':
		if (inString[pos+1]=='x' && inString[pos+2]=='p') {
			*esp=3;
      *oper=6;
		}
		else *oper=-10;
		break;
	case 'l':
		if (inString[pos+1]=='n') {
			*esp=2;
			*oper=7;
		}
		else if (inString[pos+1]=='o' && inString[pos+2]=='g'){
         	*esp=3;
          *oper=8;
         }
         else *oper=-10;
         break;
	case 's':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=17;
		}
		else if (inString[pos+1]=='q') {
			*esp=4;
			*oper=9;
		}
		else {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=17;
			}
			else {
				*esp=3;
				*oper=11;
			}
		}
		break;
	case 't':
		if (inString[pos+1]=='h') {
			*esp=2;
			*oper=19;
		}
		else if (inString[pos+1]=='a' && inString[pos+2]=='n') {
			if (inString[pos+3]=='h') {
				*esp=4;
				*oper=19;
			}
			else {
				*esp=3;
				*oper=13;
			}
		}
		else *oper=-10;
		break;
	default:
		*oper=-10;
		break;
}
}

//............................................................ ConstructTree

CPPNode * ConstructTree(char * string,short length,short & error) {
short imbric,Bimbric;
short priorite,ope;
short position,positionv,i,j;
short opetemp=0;
short espa,espat;
short caspp=0;
char * tString;
CPPNode * tCPPNode;

// Initialisation des variables

if (length<=0) {
	error=3;
	return NULL;
}

ope=0;
imbric=0;Bimbric=128;
priorite=6;
i=0;
positionv=position=0;

// Mise en place des donnees sur le morceau de chaine

while (i<length) {
	//printf("%c",string[i]);
	if ((string[i]>47 && string[i]<58) || (string[i]=='')) {
		if (priorite>5) {
			priorite=5;
			positionv=i;
		}
		i++;
	}
	else {
	if (string[i]>96 && string[i]<117) {
		FindOperator(&opetemp,&espat,string,i);
		if (opetemp>=0) {
			if (imbric<Bimbric) {
				Bimbric=imbric;
				ope=opetemp;
				position=i;
				priorite=4;
				espa=espat;
			} 
			else if (imbric==Bimbric && priorite >=4) {
				ope=opetemp;
				position=i;
				priorite=4;
				espa=espat;
			}
			j=i+1;
			i+=espat;
			while(j<i) {
				//printf("%c",string[j]);
				j++;
			}
		}
		else if (string[i]=='t') {
			if (priorite==6) ope=-1;
			i++;
		}
		else if (string[i]=='p') {
			if (priorite==6) ope=-2;
			i++;
		}
		else if (string[i]=='r') {
			if (priorite==6) ope=-2;
			i++;
		}
		else if (string[i]=='n') {
			if (priorite==6) ope=-1;
			i++;
		}
		else {
			error=2; // symbole non reconnu
      return NULL;
    }
	}
	else {
		switch(string[i]) {
			case '(':
				imbric++;
				break;
			case ')':
				imbric--;
				break;
			case '+':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=1;
					ope=1;
					position=i;
					caspp=0;
				}
				else if (imbric==Bimbric && priorite >=1) {
					priorite=1;
					ope=1;
					position=i;
					caspp=0;
				}
				break;
			case '-':
				if (imbric<Bimbric) {
					if ((i-1)<0) {
						if (priorite>5) {
							priorite=1;
							position=i;
							ope=2;
							Bimbric=imbric;
							caspp=1;
						}
					}
				 	else if (string[i-1]=='(') {
				 		if (priorite>1){
				 			priorite=1;
				 			position=i;
				 			Bimbric=imbric;
				 			caspp=1;
				 			ope=2;
				 		}
				 	}
					else {
						Bimbric=imbric;
						priorite=1;
						ope=2;
						position=i;
						caspp=0;
					}
				}
				else if (imbric==Bimbric && priorite>=1) {
					if ((i-1)<0) {
						if (priorite>5) {
							priorite=1;
							position=i;
							ope=2;
							caspp=1;
						}
					}
					else if (string[i-1]=='(') {
						if (priorite>1){
							priorite=1;
							position=i;
							caspp=1;
							ope=2;
						}
					}
					else {
						priorite=1;
						ope=2;
						position=i;
						caspp=0;
					}
				}
				break;
			case '*':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=2;
					ope=3;
					position=i;
				}
				else if (imbric==Bimbric && priorite>=2) {
					priorite=2;
					ope=3;
					position=i;
				}
				break;
			case '/':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=2;
					ope=4;
					position=i;
				}
				else if (imbric==Bimbric && priorite>=2) {
					priorite=2;
					ope=4;
					position=i;
				}
				break;
			case '^':
				if (imbric<Bimbric) {
					Bimbric=imbric;
					priorite=3;
					ope=5;
					position=i;
				}
				else if (imbric==Bimbric && priorite>=3) {
					priorite=3;
					ope=5;
					position=i;
				}
				break;
			case '.':
				break;
			case 'x':
				if (priorite==6) ope=-1;
				break;
			case 'y':
				if (priorite==6) ope=-2;
				break;
			default:
				error=2; // symbole non reconnu
				return NULL;
			}
      i++;
		}
	}
}

if (imbric!=0) {
	error=1; // erreur de parenthesage
	return NULL;
}

// Traitement des donnees	

//printf("\nMeilleure Imbrication: %d Meilleure priorite: %d Operateur: %d",Bimbric,priorite,ope);
//printf(" Position: %d Positionv: %d\n",position,positionv);
//printf("caspp: %d\n",caspp);


if (priorite==6) {
	tCPPNode =new CPPNode(ope,0);
	return tCPPNode;
}
else if (caspp==1) {
	tCPPNode = new CPPNode(2,0);
	
	tCPPNode->fg= new CPPNode(0,0);
				    
	if ((length-position-1-Bimbric)==0) { // argument absent
		error=3;
		return NULL;
	} 
	
  tString=CopyPartialString(string,(position+1),(length-1-Bimbric));
	
  tCPPNode->fd=ConstructTree(tString,(length-position-1-Bimbric),error);
	
  delete tString;
	
  return tCPPNode;
}

else if (priorite==5) {
	tCPPNode = new CPPNode(0,CalcValue(string,positionv));
	
	return tCPPNode;
}
else if (ope>5) {
	tCPPNode = new CPPNode(ope,0);
	
	if ((length-position-espa-Bimbric)==0) { // argument absent
		error=3;
		return NULL;
	}
	
  tString=CopyPartialString(string,(position+espa),(length-1));
	
  tCPPNode->fg=ConstructTree(tString,(length-position-espa-Bimbric),error);
	
  delete tString;
	
  return tCPPNode;
} 
else {
	tCPPNode = new CPPNode(ope,0);
	
	if ((position-Bimbric)==0) { // argument absent
		error=3;
		return NULL;
	}
	
  tString=CopyPartialString(string,Bimbric,(position-1));
	
  tCPPNode->fg=ConstructTree(tString,(position-Bimbric),error);
	
  delete tString;
	
  if ((length-position-1-Bimbric)==0) { // argument absent
		error=3;
		return NULL;
	}
	
  tString=CopyPartialString(string,(position+1),(length-1-Bimbric));
	
  tCPPNode->fd=ConstructTree(tString,(length-position-1-Bimbric),error);
	
  delete tString;
	
  return tCPPNode;
}
}


//...................................................... CPPNode::ComputeTree

float CPPNode::ComputeTree(float value1,float value2,short & error) {
float valueL;
float valueR;

if (mOperator==0) return mValue;
if (mOperator==-1) return value1;
if (mOperator==-2) return value2;

valueL=fg->ComputeTree(value1,value2,error);

if (error!=0) return 0;

if (fd!=NULL) valueR=fd->ComputeTree(value1,value2,error);
if (error!=0) return 0;

switch(mOperator) {
	case 1:								// +
		return (valueL+valueR);
	case 2:								// -
		return (valueL-valueR);
	case 3:								// *
		return (valueL*valueR);
	case 4:								// -
		if (valueR==0) {
			error=1;
			return 0;
		}
		return (valueL/valueR);
	case 5:								// ^
		if ((((long) valueR)-valueR)==0) return pow(valueL,valueR);
		else if (valueL<0) {
			error=2;
			return 0;
		} 
		else return pow(valueL,valueR);
		break;
	case 6:								// exp
		return exp(valueL);
	case 7:								// ln
		if (valueL<=0) {
			if (valueL<0) error=2;
			else error=1;
			return 0;
		}
		return log(valueL);
	case 8:								// log
		if (valueL<=0) {
			if (valueL<0) error=2;
			else error=1;
			return 0;
		}
		return log10(valueL);
	case 9:								// sqrt
		if (valueL<0) {
			error=2;
			return 0;
		}
		return sqrt(valueL);
	case 10:							// abs
		return fabs(valueL);
	case 11:
		return sin(valueL);			// sin
	case 12:
		return cos(valueL);			// cos
	case 13:
		return tan(valueL);			// tan
	case 14:
		return asin(valueL);			// asin
	case 15:
		return acos(valueL);			// acos
	case 16:
		return atan(valueL);			// atan
	case 17:
		return sinh(valueL);			// sinh
	case 18:
		return cosh(valueL);			// cosh
	case 19:
		return tanh(valueL);			// tanh
	default:
		return 0;
}
}

//....................................................... CPPNode::WriteTree

void CPPNode::WriteTree(char * inString) {
bool parenthese=false;

switch(mOperator) {
	case -2:
		strcat(inString,"y");
		break;
	case -1:
		strcat(inString,"x");
		break;
	case 0:
		char temp[20];
		sprintf(temp,"%f",mValue);
		strcat(inString,temp);
		break;
	case 1:
		fg->WriteTree(inString);
		strcat(inString,"+");
		fd->WriteTree(inString);
		break;
	case 2:
		if (fg->mOperator!=0 || fg->mValue!=0)
                      fg->WriteTree(inString);
		strcat(inString,"-");
		if (fd->mOperator==1 || fd->mOperator==2) {
			parenthese=true;
			strcat(inString,"(");
		}
		fd->WriteTree(inString);
		if (parenthese==true) {
			strcat(inString,")");
		}
		break;
	case 3:
		if (fg->mOperator==1 || fg->mOperator==2) {
			parenthese=true;
			strcat(inString,"(");
		}
		fg->WriteTree(inString);
		if (parenthese==true)
			strcat(inString,")");
		parenthese=false;
		strcat(inString,"*");
		if (fd->mOperator==1 || fd->mOperator==2) {
			parenthese=true;
			strcat(inString,"(");
		}
		fd->WriteTree(inString);
		if (parenthese==true)
			strcat(inString,")");
		break;
	case 4:
		if (fg->mOperator==1 || fg->mOperator==2) {
			parenthese=true;
			strcat(inString,"(");
		}
		fg->WriteTree(inString);
		if (parenthese==true)
			strcat(inString,")");
		parenthese=false;
		strcat(inString,"/");
		if (fd->mOperator>0 && fd->mOperator<5) {
			parenthese=true;
			strcat(inString,"(");
		}
		fd->WriteTree(inString);
		if (parenthese==true) 
			strcat(inString,")");
		break;
	case 5:
		if (fg->mOperator>0 && fg->mOperator<5) {
			parenthese=true;
			strcat(inString,"(");
		}
		fg->WriteTree(inString);
		if (parenthese==true)
			strcat(inString,")");
		parenthese=false;
		strcat(inString,"^");
		if (fd->mOperator>0 && fd->mOperator<5) {
			parenthese=true;
			strcat(inString,"(");
		}
		fd->WriteTree(inString);
		if (parenthese==true) 
			strcat(inString,")");
		break;
	case 6:
		strcat(inString,"exp(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 7:
		strcat(inString,"log(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 8:
		strcat(inString,"ln(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 9:
		strcat(inString,"sqrt(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 10:
		strcat(inString,"|");
		fg->WriteTree(inString);
		strcat(inString,"|");
		break;
	case 11:
		strcat(inString,"sin(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 12:
		strcat(inString,"cos(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 13:
		strcat(inString,"tan(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 14:
		strcat(inString,"asin(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 15:
		strcat(inString,"acos(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 16:
		strcat(inString,"atan(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 17:
		strcat(inString,"sinh(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 18:
		strcat(inString,"cosh(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
	case 19:
		strcat(inString,"tanh(");
		fg->WriteTree(inString);
		strcat(inString,")");
		break;
}
}

//...................................................... CPPNode::ObjetWidth

/*short CPPNode::ObjetWidth(void) {
// Permet de determiner la largeur en pixels d'un operateur ou scalaire
// A vous de voir ce que vous voulez

return ESPACE_LARGEUR;
}*/

//...................................................... CPPNode::DrawContent

/*void CPPNode::DrawContent(Point pt) {
short longueur=ObjetWidth();
Rect r;

::SetRect(&r,pt.h,pt.v,pt.h+longueur,pt.v+HAUTEUR);
::FrameRect(&r);

// A completer comme bon vous semble par un DrawString,
// TextBox, ou encore  UDrawingUtils (PowerPlant) 

}*/

//....................................................... CPPNode::Draw_Tree

/*Point CPPNode::Draw_Tree(Point & pt1) {
Point ptd,ptd1,ptd2;
short decalage=ESPACE_LARGEUR+ObjetWidth();

if (fg!=NULL) {
	pt1.h+=decalage;
	ptd1=fg->Draw_Tree(pt1);
	pt1.h-=decalage;
	if (fd!=NULL) {
		pt1.h+=decalage;
		ptd2=fd->Draw_Tree(pt1);
		pt1.h-=decalage;
		ptd.h=pt1.h;
		ptd.v=(ptd1.v+ptd2.v)/2;
	}
	else {
		ptd.h=pt1.h;
		ptd.v=ptd1.v;
	}
}
else {
	ptd=pt1;
	ptd.v+=ESPACE_HAUTEUR;
	pt1.v+=ESPACE_HAUTEUR+HAUTEUR;
}
if (fg!=NULL) {
	::MoveTo(ptd1.h,ptd1.v+(HAUTEUR/2));
	::LineTo(ptd.h+ObjetWidth(),ptd.v+(HAUTEUR/2));
	if (fd!=NULL) {
		::MoveTo(ptd2.h,ptd2.v+(HAUTEUR/2));
		::LineTo(ptd.h+ObjetWidth(),ptd.v+(HAUTEUR/2));
	}
}

DrawContent(ptd);

return ptd;
}*/

#pragma mark -

//....................................................... CPPNode::operator =

CPPNode & CPPNode::operator =(const CPPNode & inCPPNode) {
mOperator=inCPPNode.mOperator;
mValue=inCPPNode.mValue;

if (inCPPNode.fg!=NULL) {
	fg=new CPPNode;
	fg=inCPPNode.fg;
}

if (inCPPNode.fd!=NULL) {
	fd=new CPPNode;
	fd=inCPPNode.fd;
}

return (*this);
}

//............................................................... operator ==

bool operator == (const CPPNode & CPPNode1 ,const CPPNode & CPPNode2) {
if (CPPNode1.mOperator!=CPPNode2.mOperator || CPPNode1.mValue!=CPPNode2.mValue) return false;
if (CPPNode1.fg==CPPNode2.fg && CPPNode1.fd==CPPNode2.fd) return true;
return false;
}

//............................................................... operator !=

bool operator != (const CPPNode & CPPNode1 ,const CPPNode & CPPNode2) {
if (CPPNode1.mOperator!=CPPNode2.mOperator || CPPNode1.mValue!=CPPNode2.mValue) return false;
if (CPPNode1.fg==CPPNode2.fg && CPPNode1.fd==CPPNode2.fd) return false;
return true;
}

//............................................................... ToolBoxInit

/* Valable uniquement sous MacOS */

/*void ToolBoxInit( void ) {

long gestaltResponse;
	
InitGraf(&qd.thePort);
InitFonts();
InitWindows();
InitMenus();
TEInit();
InitDialogs(0L);
InitCursor();
	
MaxApplZone();
}*/

/*************************************************************************/

/* A commenter pour utilisation dans un autre projet */

int main(int argc, char ** argv) {
char tString2[500];
char tString[256];
short  err=0;

CPPNode * test;

//Rect r={0,0,800,600};         MacOS Uniquement
//Point pt={50,50};             MacOS Uniquement

//ToolBoxInit();                MacOS Uniquement

printf("Entrez votre expression parenthesee: \n");
scanf("%s",tString);
   
printf("%d\n",strlen(tString));

test=ConstructTree(tString,strlen(tString),err);
printf("Type d'erreur: %d\n", err);

if (err==0) {
	printf("Evaluation: %f\n",test->ComputeTree(1,1,err));
	printf("Type d'erreur: %d\n", err);
	*tString2='\0';
	test->WriteTree((char *)tString2);
	printf("%s\n",tString);
	printf("%s\n",tString2);
}

//WindowPtr window= NewCWindow(nil,&r,"\pTest",true,0,(WindowPtr) -1,true,0);

//SelectWindow(window);         MacOS Uniquement
//SetPort(window);              MacOS Uniquement
//test->Draw_Tree(pt);          MacOS Uniquement
return 0;
}
